/*++

Copyright (c) 1992 Microsoft Corporation

Module Name:

    SockProc.c

Abstract:

    This module contains support routines for the WinSock DLL.

Author:

    David Treadwell (davidtr)    20-Feb-1992

Revision History:

--*/

#include "winsockp.h"

#include <ctype.h>
#include <stdarg.h>
#include <wincon.h>


VOID
SockBuildSockaddr (
    OUT PSOCKADDR Sockaddr,
    OUT PINT SockaddrLength,
    IN PTRANSPORT_ADDRESS TdiAddress
    )
{
    WS_ASSERT( sizeof(TdiAddress->Address[0].AddressType) ==
                sizeof(Sockaddr->sa_family) );
    WS_ASSERT( FIELD_OFFSET( TA_ADDRESS, AddressLength ) == 0 );
    WS_ASSERT( FIELD_OFFSET( TA_ADDRESS, AddressType ) == sizeof(USHORT) );
    WS_ASSERT( FIELD_OFFSET( TRANSPORT_ADDRESS, Address[0] ) == sizeof(int) );
    WS_ASSERT( FIELD_OFFSET( SOCKADDR, sa_family ) == 0 );

    //
    // Convert the specified TDI address to a sockaddr.
    //

    *SockaddrLength = TdiAddress->Address[0].AddressLength +
                          sizeof(Sockaddr->sa_family);

    RtlCopyMemory(
        Sockaddr,
        &TdiAddress->Address[0].AddressType,
        *SockaddrLength
        );

    return;

} // SockBuildSockaddr


VOID
SockBuildTdiAddress (
    OUT PTRANSPORT_ADDRESS TdiAddress,
    IN PSOCKADDR Sockaddr,
    IN INT SockaddrLength
    )
{
    WS_ASSERT( sizeof(TdiAddress->Address[0].AddressType) ==
                sizeof(Sockaddr->sa_family) );
    WS_ASSERT( FIELD_OFFSET( TA_ADDRESS, AddressLength ) == 0 );
    WS_ASSERT( FIELD_OFFSET( TA_ADDRESS, AddressType ) == sizeof(USHORT) );
    WS_ASSERT( FIELD_OFFSET( TRANSPORT_ADDRESS, Address[0] ) == sizeof(int) );
    WS_ASSERT( FIELD_OFFSET( SOCKADDR, sa_family ) == 0 );

    //
    // Convert the specified sockaddr to a TDI address.
    //

    TdiAddress->TAAddressCount = 1;
    TdiAddress->Address[0].AddressLength =
        SockaddrLength - sizeof(Sockaddr->sa_family) ;

    RtlCopyMemory(
        &TdiAddress->Address[0].AddressType,
        Sockaddr,
        SockaddrLength
        );

    return;

} // SockBuildTdiAddress


VOID
SockDereferenceSocket (
    IN PSOCKET_INFORMATION Socket
    )

/*++

Routine Description:

    Dereferences the specified socket and, if necessary, removes it from
    the global list of sockets.

Arguments:

    Socket - a pointer to the socket to dereference.

Return Value:

    None.

--*/

{
    WS_ASSERT( Socket->ReferenceCount > 0 );

    //
    // Acquire the resource that protects the socket reference count.
    //

    SockAcquireGlobalLockExclusive( );

    //
    // Dereference the socket.
    //

    Socket->ReferenceCount--;

    //
    // If the reference count on the socket has dropped to zero, then
    // nobody is accessing it and the active reference has gone away,
    // so it is time to delete the socket from the process's list
    // of sockets.
    //

    if ( Socket->ReferenceCount == 0 ) {

        //
        // Delete the resource on the socket.
        //

        RtlDeleteResource( &Socket->Lock );

        //
        // Free the local and remote addresses on the socket.
        //

        if ( Socket->LocalAddress != NULL ) {
            FREE_HEAP( Socket->LocalAddress );
        }

        if ( Socket->RemoteAddress != NULL ) {
            FREE_HEAP( Socket->RemoteAddress );
        }

        //
        // Remove the socket from the global list of sockets.
        //

        RemoveEntryList( &Socket->SocketListEntry );

        SockReleaseGlobalLock( );

        //
        // Finally, free the socket structure itself.
        //

        FREE_HEAP( Socket );

    } else {

        SockReleaseGlobalLock( );
    }

    return;

} // SockDereferenceSocket


PSOCKET_INFORMATION
SockFindAndReferenceSocket (
    IN SOCKET Handle,
    IN BOOLEAN AttemptImport
    )

/*++

Routine Description:

    Looks up a socket in the global socket table, and references
    it if found.

Arguments:

    Handle - NT system handle of the socket to locate.

    AttemptImport - if the socket isn't currently valid in this
        process, this parameter specifies whether we should attempt
        to import the handle into this process.

Return Value:

    PSOCKET_INFORMATION - a referenced pointer to a socket structure,
        or NULL if none was found that matched the specified handle.

--*/

{
    PSOCKET_INFORMATION socket;
    PLIST_ENTRY listEntry;
    BOOLEAN found = FALSE;

    //
    // Acquire the resource that protects sockets.
    //

    SockAcquireGlobalLockExclusive( );

    //
    // Attempt to find the socket in the list.
    //

    for ( listEntry = SocketListHead.Flink;
          listEntry != &SocketListHead;
          listEntry = listEntry->Flink ) {

        socket = CONTAINING_RECORD(
                     listEntry,
                     SOCKET_INFORMATION,
                     SocketListEntry
                     );

        if ( socket->Handle == Handle && socket->State != SocketStateClosing ) {
            found = TRUE;
            break;
        }
    }

    //
    // If the socket wasn't found, check to see if AFD knows about the
    // handle.  If AFD knows about it, then it was either inherited or
    // duped into this process, and we need to set up state for it.
    //

    if ( !found ) {

        if ( AttemptImport ) {

            socket = SockGetHandleContext( Handle );
    
            if ( socket == NULL ) {
                SockReleaseGlobalLock( );
                return NULL;
            }

        } else {

            SockReleaseGlobalLock( );
            return NULL;
        }
    }

    //
    // The socket was found, so reference the socket and return.  The
    // reference ensures that the socket information structure will not
    // be deallocated while somebody is looking at it.  However, it is
    // the responsibility of the caller to dereference the socket when
    // it is done using the structure.
    //

    WS_ASSERT( socket->ReferenceCount > 0 );

    socket->ReferenceCount++;

    SockReleaseGlobalLock( );

    return socket;

} // SockFindAndReferenceSocket


PSOCKET_INFORMATION
SockGetHandleContext (
    IN SOCKET Handle
    )
{
    NTSTATUS status;
    IO_STATUS_BLOCK ioStatusBlock;
    ULONG contextLength;
    PVOID context;
    PSOCKET_INFORMATION newSocket;
    UNICODE_STRING transportDeviceName;
    BOOLEAN succeeded;
    ULONG helperDllContextLength;
    PCHAR contextPtr;
    INT error;
    BOOLEAN resourceInitialized;

    //
    // Get the lock that protects access to socket lists, etc.
    //

    SockAcquireGlobalLockExclusive( );

    //
    // Initialize locals so that we know how to clean up on exit.
    //

    context = NULL;
    newSocket = NULL;
    succeeded = FALSE;
    resourceInitialized = FALSE;

    //
    // Call AFD to determine the length of context info for the socket.
    // If this succeeds, then it is most likely true that the handle
    // is valid for this process.
    //

    status = NtDeviceIoControlFile(
                 (HANDLE)Handle,
                 SockThreadEvent,
                 NULL,                   // APC Routine
                 NULL,                   // APC Context
                 &ioStatusBlock,
                 IOCTL_AFD_GET_CONTEXT_LENGTH,
                 NULL,
                 0,
                 &contextLength,
                 sizeof(contextLength)
                 );



    if ( status == STATUS_PENDING ) {
        SockWaitForSingleObject(
            SockThreadEvent,
            Handle,
            SOCK_NEVER_CALL_BLOCKING_HOOK,
            SOCK_NO_TIMEOUT
            );
        status = ioStatusBlock.Status;
    }

    if ( !NT_SUCCESS(status) || contextLength < sizeof(*newSocket) ) {
        goto exit;
    }

    //
    // Now allocate memory to hold the socket context and get the actual
    // context for the socket.
    //

    context = ALLOCATE_HEAP( contextLength );
    if ( context == NULL ) {
        goto exit;
    }

    status = NtDeviceIoControlFile(
                 (HANDLE)Handle,
                 SockThreadEvent,
                 NULL,                   // APC Routine
                 NULL,                   // APC Context
                 &ioStatusBlock,
                 IOCTL_AFD_GET_CONTEXT,
                 NULL,
                 0,
                 context,
                 contextLength
                 );

    if ( status == STATUS_PENDING ) {
        SockWaitForSingleObject(
            SockThreadEvent,
            Handle, 
            SOCK_NEVER_CALL_BLOCKING_HOOK,
            SOCK_NO_TIMEOUT
            );
        status = ioStatusBlock.Status;
    }

    if ( !NT_SUCCESS(status) ) {
        goto exit;
    }

    //
    // We have obtained the necessary context for the socket.  The context
    // information is structured as follows:
    //
    //     SOCKET_INFORMATION structure
    //     Helper DLL Context Length
    //     Local Address
    //     Remote Address
    //     Helper DLL Context
    //
    // First allocate a socket information structure for this socket.
    //

    newSocket = ALLOCATE_HEAP( sizeof(*newSocket) );
    if ( newSocket == NULL ) {
        goto exit;
    }

    //
    // Copy in to the new socket information structure the initial context.
    //

    RtlCopyMemory( newSocket, context, sizeof(*newSocket) );

    //
    // Initialize various fields in the socket information structure.
    //

    newSocket->Handle = Handle;
    newSocket->ReferenceCount = 2;
    newSocket->ConnectInProgress = FALSE;
    newSocket->TdiAddressHandle = NULL;
    newSocket->TdiConnectionHandle = NULL;
    newSocket->HelperDll = NULL;
    newSocket->SocketListEntry.Flink = NULL;
    newSocket->LocalAddress = NULL;
    newSocket->RemoteAddress = NULL;

    try {
        RtlInitializeResource( &newSocket->Lock );
    } except ( EXCEPTION_EXECUTE_HANDLER ) {
        goto exit;
    }

    resourceInitialized = TRUE;

    //
    // Determine the length of the helper DLL's context information.
    //

    contextPtr = (PCHAR)context + sizeof(*newSocket);
    helperDllContextLength = *(PULONG)contextPtr;
    contextPtr += sizeof(ULONG);

    //
    // Get the helper DLL for the socket loaded.
    //

    error = SockGetTdiName(
                &newSocket->AddressFamily,
                &newSocket->SocketType,
                &newSocket->Protocol,
                &transportDeviceName,
                &newSocket->HelperDllContext,
                &newSocket->HelperDll,
                &newSocket->HelperDllNotificationEvents
                );
    if ( error != NO_ERROR ) {
        goto exit;
    }

    //
    // Allocate buffers for the local and remote addresses and copy in
    // information from the context buffer retrieved from AFD.
    //

    WS_ASSERT( newSocket->HelperDll != NULL );

    newSocket->LocalAddress =
        ALLOCATE_HEAP( newSocket->HelperDll->MaxSockaddrLength );
    if ( newSocket->LocalAddress == NULL ) {
        goto exit;
    }
    newSocket->LocalAddressLength = newSocket->HelperDll->MaxSockaddrLength;

    RtlCopyMemory(
        newSocket->LocalAddress,
        contextPtr,
        newSocket->LocalAddressLength
        );
    contextPtr += newSocket->LocalAddressLength;

    newSocket->RemoteAddress =
        ALLOCATE_HEAP( newSocket->HelperDll->MaxSockaddrLength );
    if ( newSocket->RemoteAddress == NULL ) {
        goto exit;
    }
    newSocket->RemoteAddressLength = newSocket->HelperDll->MaxSockaddrLength;

    RtlCopyMemory(
        newSocket->RemoteAddress,
        contextPtr,
        newSocket->RemoteAddressLength
        );
    contextPtr += newSocket->RemoteAddressLength;

    //
    // Get TDI handles for this socket.
    //

    error = SockGetTdiHandles( newSocket );
    if ( error != NO_ERROR ) {
        goto exit;
    }

    //
    // Inform the helper DLL about this new socket.
    //

    error = newSocket->HelperDll->WSHSetSocketInformation (
                NULL,
                newSocket->Handle,
                newSocket->TdiAddressHandle,
                newSocket->TdiConnectionHandle,
                SOL_INTERNAL,
                SO_CONTEXT,
                contextPtr,
                helperDllContextLength
                );
    if ( error != NO_ERROR ) {
        goto exit;
    }

    //
    // Remember the helper DLL socket context value.
    //

    newSocket->HelperDllContext = *(PVOID *)contextPtr;

    //
    // Give the socket a unique serial number.  This is used to identify
    // the socket for AsyncSelect requests.
    //

    newSocket->SocketSerialNumber = SockSocketSerialNumberCounter++;

    //
    // Place the socket information structure in the process's global
    // list of sockets.
    //

    InsertHeadList( &SocketListHead, &newSocket->SocketListEntry );

    //
    // If the socket has AsyncSelect events set up, set them up for this
    // process.
    //

    if ( newSocket->AsyncSelectlEvent ) {

        error = WSAAsyncSelect(
                    newSocket->Handle,
                    newSocket->AsyncSelecthWnd,
                    newSocket->AsyncSelectwMsg,
                    newSocket->AsyncSelectlEvent
                    );
        if ( error == SOCKET_ERROR ) {
            goto exit;
        }
    }

    succeeded = TRUE;

exit:

    if ( !succeeded && newSocket != NULL ) {

        if ( resourceInitialized ) {
            RtlDeleteResource( &newSocket->Lock );
        }

        if ( newSocket->TdiAddressHandle != NULL ) {
            status = NtClose( newSocket->TdiAddressHandle );
            WS_ASSERT( NT_SUCCESS(status) );
        }

        if ( newSocket->TdiConnectionHandle != NULL ) {
            status = NtClose( newSocket->TdiConnectionHandle );
            WS_ASSERT( NT_SUCCESS(status) );
        }

        if ( newSocket->LocalAddress != NULL ) {
            FREE_HEAP( newSocket->LocalAddress );
        }

        if ( newSocket->RemoteAddress != NULL ) {
            FREE_HEAP( newSocket->RemoteAddress );
        }

        if ( newSocket->HelperDll != NULL ) {
            SockNotifyHelperDll( newSocket, WSH_NOTIFY_CLOSE );
        }

        if ( newSocket->SocketListEntry.Flink != NULL ) {
            RemoveEntryList( &newSocket->SocketListEntry );
        }

        FREE_HEAP( newSocket );

        newSocket = NULL;

        IF_DEBUG(SOCKET) {
            WS_PRINT(( "SockGetHandleContext: failed to import socket "
                       "handle %lx\n", Handle ));
        }

    } else if ( !succeeded ) {

        IF_DEBUG(SOCKET) {
            WS_PRINT(( "SockGetHandleContext: failed to import socket "
                       "handle %lx, unknown to AFD: %lX\n", Handle, status ));
        }

    } else {

        WS_ASSERT( succeeded );

        IF_DEBUG(SOCKET) {
            WS_PRINT(( "Imported socket %lx (%lx) of type %s\n",
                           newSocket->Handle, &newSocket,
                           (newSocket->SocketType == SOCK_DGRAM ?
                                             "SOCK_DGRAM" : "SOCK_STREAM") ));
        }

        SockDereferenceSocket( newSocket );
    }

    if ( context != NULL ) {
        FREE_HEAP( context );
    }

    SockReleaseGlobalLock( );

    return newSocket;

} // SockGetHandleContext


INT
SockSetHandleContext (
    IN PSOCKET_INFORMATION Socket
    )
{
    NTSTATUS status;
    PVOID context;
    PCHAR contextPtr;
    ULONG contextLength;
    ULONG helperDllContextLength;
    IO_STATUS_BLOCK ioStatusBlock;
    INT error;

    //
    // Determine how much space we need for the helper DLL context.
    //

    error = Socket->HelperDll->WSHGetSocketInformation (
                Socket->HelperDllContext,
                Socket->Handle,
                Socket->TdiAddressHandle,
                Socket->TdiConnectionHandle,
                SOL_INTERNAL,
                SO_CONTEXT,
                NULL,
                (PINT)&helperDllContextLength
                );
    if ( error != NO_ERROR ) {
        return NO_ERROR;  // !!!
        //return error;
    }

    //
    // Allocate a buffer to hold all context information.
    //

    contextLength = sizeof(*Socket) + Socket->LocalAddressLength +
                        Socket->RemoteAddressLength +
                        sizeof(helperDllContextLength) + helperDllContextLength;

    context = ALLOCATE_HEAP( contextLength );
    if ( context == NULL ) {
        error = WSAENOBUFS;
        return error;
    }

    //
    // Copy over information to the context buffer.  The context buffer
    // has the following format:
    //
    //     SOCKET_INFORMATION structure
    //     Helper DLL Context Length
    //     Local Address
    //     Remote Address
    //     Helper DLL Context
    //

    contextPtr = context;

    RtlCopyMemory( contextPtr, Socket, sizeof(*Socket) );
    contextPtr += sizeof(*Socket);

    *(PULONG)contextPtr = helperDllContextLength;
    contextPtr += sizeof(helperDllContextLength);

    RtlCopyMemory(contextPtr, Socket->LocalAddress, Socket->LocalAddressLength );
    contextPtr += Socket->LocalAddressLength;

    RtlCopyMemory(contextPtr, Socket->RemoteAddress, Socket->RemoteAddressLength );
    contextPtr += Socket->RemoteAddressLength;

    //
    // Get the context from the helper DLL.
    //

    error = Socket->HelperDll->WSHGetSocketInformation (
                Socket->HelperDllContext,
                Socket->Handle,
                Socket->TdiAddressHandle,
                Socket->TdiConnectionHandle,
                SOL_INTERNAL,
                SO_CONTEXT,
                contextPtr,
                (PINT)&helperDllContextLength
                );
    if ( error != NO_ERROR ) {
        FREE_HEAP( context );
        return error;
    }

    //
    // Now give all this information to AFD to hold on to.
    //

    status = NtDeviceIoControlFile(
                 (HANDLE)Socket->Handle,
                 SockThreadEvent,
                 NULL,                   // APC Routine
                 NULL,                   // APC Context
                 &ioStatusBlock,
                 IOCTL_AFD_SET_CONTEXT,
                 context,
                 contextLength,
                 NULL,
                 0
                 );

    if ( status == STATUS_PENDING ) {
        SockWaitForSingleObject(
            SockThreadEvent,
            Socket->Handle,
            SOCK_NEVER_CALL_BLOCKING_HOOK,
            SOCK_NO_TIMEOUT
            );
        status = ioStatusBlock.Status;
    }

    FREE_HEAP( context );

    if ( !NT_SUCCESS(status) ) {
        error = SockNtStatusToSocketError( status );
        return error;
    }

    return NO_ERROR;

} // SockSetHandleContext


VOID
SockReferenceSocket (
    IN PSOCKET_INFORMATION Socket
    )

/*++

Routine Description:

    Increments the reference count on the passed-in socket handle.

Arguments:

    Socket - points to the socket information structure.

Return Value:

    PSOCKET_INFORMATION - a referenced pointer to a socket structure,
        or NULL if none was found that matched the specified handle.

--*/

{
    //
    // Acquire the resource that protects socket reference counts.
    //

    SockAcquireGlobalLockExclusive( );

    //
    // The reference ensures that the socket information structure will
    // not be deallocated while somebody is looking at it.  However, it
    // is the responsibility of the caller to dereference the socket
    // when it is done using the structure.
    //

    WS_ASSERT( Socket->ReferenceCount > 0 );

    Socket->ReferenceCount++;

    SockReleaseGlobalLock( );

    return;

} // SockReferenceSocket


BOOL
SockWaitForSingleObject (
    IN HANDLE Handle,
    IN SOCKET SocketHandle,
    IN DWORD BlockingHookUsage,
    IN DWORD TimeoutUsage
    )

/*++

Routine Description:

    Does an alertable wait on the specified handle.  If the wait completes
    due to an alert, it rewaits.

Arguments:

    Handle - NT system handle to wait on.

    SocketHandle - the socket handle on which we're performing the IO
        we're waiting for.  This is necessary to support
        WSACancelBlockingCall().

    BlockingHookUsage - indicates whether to call the thread's blocking
        hook.  Possible values are:

            SOCK_ALWAYS_CALL_BLOCKING_HOOK - blocking hook is always
                called if blocking is necessary.
                
            SOCK_CONDITIONALLY_CALL_BLOCKING_HOOK - blocking hook is
                called if the socket is blocking (i.e. not a nonblocking
                socket).
               
            SOCK_NEVER_CALL_BLOCKING_HOOK - blocking hook is never
                called.

    TimeoutUsage - determines whether to wait infinitely or for a
        timeout.  Possible values are:

            SOCK_NO_TIMEOUT - wait forever for the handle to be
                signalled.

            SOCK_SEND_TIMEOUT - use the socket's send timeout value
                as a timeout.
                
            SOCK_RECEIVE_TIMEOUT - use the socket's receive timeout
                value as a timeout.

Return Value:

    BOOL - TRUE if the object was signalled within the appropriate
        timeout, and FALSE if the timeout occurred first.

--*/

{
    NTSTATUS status;
    LARGE_INTEGER timeout;
    BOOLEAN callBlockingHook;
    BOOLEAN useTimeout;
    LARGE_INTEGER endTime;
    LARGE_INTEGER currentTime;
    PSOCKET_INFORMATION socket = NULL;

    //
    // First wait for the object for a little while.  This handles the 
    // usual case where the object is already signalled or is signalled 
    // shortly into the wait.  We'll only go through the longer, more 
    // complex path if we're going to have to wait longer.  
    //

    timeout.HighPart = 0xFFFFFFFF;
    timeout.LowPart = (ULONG)(-1 * (10*1000*500));     // 0.5 seconds

    status = NtWaitForSingleObject( Handle, TRUE, &timeout );
    if ( status == STATUS_SUCCESS ) {
        return TRUE;
    }

    //
    // If we need to extract information from the socket, get a pointer 
    // to the socket information structure.  
    //

    if ( BlockingHookUsage == SOCK_CONDITIONALLY_CALL_BLOCKING_HOOK ||
             TimeoutUsage == SOCK_SEND_TIMEOUT ||
             TimeoutUsage == SOCK_RECEIVE_TIMEOUT ) {

        socket = SockFindAndReferenceSocket( SocketHandle, FALSE );
        if ( socket == NULL ) {
            NtWaitForSingleObject( Handle, TRUE, NULL );
            return TRUE;
        }
    }

    //
    // Determine whether we need to call the blocking hook while
    // we're waiting.
    //

    switch ( BlockingHookUsage ) {

    case SOCK_ALWAYS_CALL_BLOCKING_HOOK:

        //
        // If we're supposed to call the blocking hook but the thread
        // has no blocking hook installed, then there is no need to call
        // the blocking hook.
        //

        if ( SockThreadBlockingHook == NULL ) {
            callBlockingHook = FALSE;
        } else {
            callBlockingHook = TRUE;
        }

        break;

    case SOCK_CONDITIONALLY_CALL_BLOCKING_HOOK:

        if ( socket->NonBlocking || SockThreadBlockingHook == NULL ) {
            callBlockingHook = FALSE;
        } else {
            callBlockingHook = TRUE;
        }

        break;

    case SOCK_NEVER_CALL_BLOCKING_HOOK:

        callBlockingHook = FALSE;
        break;

    default:

        WS_ASSERT( FALSE );
        break;
    }

    //
    // Determine what our timeout should be, if any.
    //

    switch ( TimeoutUsage ) {

    case SOCK_NO_TIMEOUT:

        useTimeout = FALSE;
        break;

    case SOCK_SEND_TIMEOUT:

        if ( socket->SendTimeout != 0 ) {
            useTimeout = TRUE;
            timeout = RtlEnlargedIntegerMultiply( socket->SendTimeout, 10*1000 );
        } else {
            useTimeout = FALSE;
        }

        break;

    case SOCK_RECEIVE_TIMEOUT:

        if ( socket->ReceiveTimeout != 0 ) {
            useTimeout = TRUE;
            timeout = RtlEnlargedIntegerMultiply( socket->ReceiveTimeout, 10*1000 );
        } else {
            useTimeout = FALSE;
        }

        break;

    default:

        WS_ASSERT( FALSE );
        break;
    }

    //
    // Dereference the socket if we got a pointer to the socket 
    // information structure.  
    //

    if ( socket != NULL ) {
        SockDereferenceSocket( socket );
#if DBG
        socket = NULL;
#endif
    }

    //
    // Calculate the end time we'll use when waiting on the handle.  The 
    // end time is the time at which we must quit waiting on the handle 
    // and must instead return from this function. 
    //

    if ( useTimeout ) {

        //
        // The end time if the current time plus the timeout.  Query
        // the current time.
        //

        status = NtQuerySystemTime( &currentTime );
        WS_ASSERT( NT_SUCCESS(status) );

        endTime = RtlLargeIntegerAdd( currentTime, timeout );

    } else {

        //
        // We need an infinite timeout.  Set the end time to the largest 
        // possible time in NT format.  
        //

        endTime.LowPart = 0xFFFFFFFF;
        endTime.HighPart = 0x7FFFFFFF;
    }

    //
    // If we're going to be calling a blocking hook, set up a minimal 
    // timeout since we have to call the blocking hook instead of idly 
    // waiting.  If we won't be calling the blocking hook, then we'll
    // wait until the end time.
    //

    if ( callBlockingHook ) {
        timeout.LowPart = 0xFFFFFFFF;
        timeout.HighPart = 0xFFFFFFFF;
    } else {
        timeout = endTime;
    }

    //
    // Remember that we're in a blocking call to prevent other winsock
    // calls from succeeding.  Also initialize the thread's cancel
    // Boolean so that we can tell whether the IO has been cancelled,
    // and remember the socket handle on which we're doing the IO.
    //

    WS_ASSERT( !SockThreadIsBlocking );

    SockThreadIsBlocking = TRUE;
    SockThreadSocketHandle = SocketHandle;

    do {

        //
        // If necessary, call the blocking hook function until it
        // returns FALSE.  This gives the routine the oppurtunity
        // to process all the available messages before we complete
        // the wait.
        //

        if ( callBlockingHook ) {
            ASSERT( SockThreadBlockingHook != NULL );
            while( (SockThreadBlockingHook( ) & 0xFF) != 0 );
        }

        //
        // If the operation was cancelled, reset the timeout to infinite
        // and wait for the cancellation.  We don't want to call the
        // blocking hook after the IO is cancelled.
        //

        if ( SockThreadIoCancelled ) {

            timeout.LowPart = 0xFFFFFFFF;
            timeout.HighPart = 0x7FFFFFFF;

        } else {

            //
            // Determine whether we have exceeded the end time.  If we 
            // have exceeded the end time then we must not wait any 
            // longer.  
            //

            status = NtQuerySystemTime( &currentTime );
            WS_ASSERT( NT_SUCCESS(status) );

            if ( RtlLargeIntegerGreaterThan( currentTime, endTime ) ) {
                status = STATUS_TIMEOUT;
                break;
            }
        }

        //
        // Perform the actual wait on the object handle.
        //

        status = NtWaitForSingleObject( Handle, TRUE, &timeout );
        WS_ASSERT( NT_SUCCESS(status) );
        WS_ASSERT( status != STATUS_TIMEOUT || !SockThreadIoCancelled );

    } while ( status == STATUS_USER_APC ||
              status == STATUS_ALERTED ||
              status == STATUS_TIMEOUT );

    //
    // Reset thread variables.
    //

    SockThreadIsBlocking = FALSE;
    SockThreadSocketHandle = INVALID_SOCKET;

    //
    // Return TRUE if the wait's return code was success; otherwise, we
    // had to timeout the wait so return FALSE.
    //

    if ( status == STATUS_SUCCESS ) {
        return TRUE;
    } else {
        return FALSE;
    }

} // SockWaitForSingleObject


BOOL
SockDefaultBlockingHook (
    VOID
    )
{

    MSG msg;
    BOOLEAN retrievedMessage;

    //
    // Get the next message for this thread, if any.
    //

    retrievedMessage = PeekMessage( &msg, NULL, 0, 0, PM_REMOVE );

    //
    // Process the message if we got one.
    //

    if ( retrievedMessage ) {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
    }

    //
    // If we got a message, indicate that we want to be called again.
    //

    return retrievedMessage;

} // SockDefaultBlockingHook


BOOLEAN
SockIsSocketConnected (
    IN PSOCKET_INFORMATION Socket
    )
{
    NTSTATUS status;

    //
    // If there is a connect in progress, call NtTestAlert() to
    // give the connect completion APC a chance to run.
    //
    // !!! Note that if this routine is called in a different thread from
    //     the one that initiated the connect, it is possible that the
    //     connect has completed and the APC hasn't run, which would result
    //     in this routine giving an incorrect answer.  Do we care about
    //     this?

    if ( Socket->ConnectInProgress ) {

        do {
            status = NtTestAlert( );
        } while ( status == STATUS_ALERTED || status == STATUS_USER_APC );
    }

    //
    // Check whether the socket is already connected.
    //

    if ( Socket->State == SocketStateConnected ) {
        return TRUE;
    }

    return FALSE;

} // SockIsSocketConnected


INT
SockGetTdiHandles (
    IN PSOCKET_INFORMATION Socket
    )
{
    AFD_HANDLE_INFO handleInfo;
    NTSTATUS status;
    IO_STATUS_BLOCK ioStatusBlock;
    ULONG getHandleInfo;

    //
    // Determine which handles we need to get.
    //

    getHandleInfo = 0;

    if ( Socket->TdiAddressHandle == NULL ) {
        getHandleInfo |= AFD_QUERY_ADDRESS_HANDLE;
    }

    if ( Socket->TdiConnectionHandle == NULL ) {
        getHandleInfo |= AFD_QUERY_CONNECTION_HANDLE;
    }

    //
    // If we already have both TDI handles for the socket, just return.
    //

    if ( getHandleInfo == 0 ) {
        return NO_ERROR;
    }

    //
    // Call AFD to retrieve the TDI handles for the socket.
    //

    status = NtDeviceIoControlFile(
                 (HANDLE)Socket->Handle,
                 SockThreadEvent,
                 NULL,                   // APC Routine
                 NULL,                   // APC Context
                 &ioStatusBlock,
                 IOCTL_AFD_QUERY_HANDLES,
                 &getHandleInfo,
                 sizeof(getHandleInfo),
                 &handleInfo,
                 sizeof(handleInfo)
                 );

    // *** Because this routine can be called at APC level from
    //     ConnectCompletionApc(), IOCTL_AFD_QUERY_HANDLES must
    //     never pend.

    WS_ASSERT( status != STATUS_PENDING );

    if ( status == STATUS_PENDING ) {
        SockWaitForSingleObject(
            SockThreadEvent,
            Socket->Handle, 
            SOCK_NEVER_CALL_BLOCKING_HOOK,
            SOCK_NO_TIMEOUT
            );
        status = ioStatusBlock.Status;
    }

    if ( !NT_SUCCESS(status) ) {
        return SockNtStatusToSocketError( status );
    }

    //
    // Set up the handles that we were returned.
    //

    if ( Socket->TdiAddressHandle == NULL ) {
        Socket->TdiAddressHandle = handleInfo.TdiAddressHandle;
    }

    if ( Socket->TdiConnectionHandle == NULL ) {
        Socket->TdiConnectionHandle = handleInfo.TdiConnectionHandle;
    }

    return NO_ERROR;

} // SockGetTdiHandles


INT
SockGetInformation (
    IN PSOCKET_INFORMATION Socket,
    IN ULONG InformationType,
    IN PVOID AdditionalInputInfo OPTIONAL,
    IN ULONG AdditionalInputInfoLength,
    IN OUT PBOOLEAN Boolean OPTIONAL,
    IN OUT PULONG Ulong OPTIONAL,
    IN OUT PLARGE_INTEGER LargeInteger OPTIONAL
    )
{
    NTSTATUS status;
    IO_STATUS_BLOCK ioStatusBlock;
    PAFD_INFORMATION afdInfo;
    ULONG afdInfoLength;

    //
    // Allocate space for the I/O buffer.
    //

    afdInfoLength = sizeof(*afdInfo) + AdditionalInputInfoLength;
    afdInfo = ALLOCATE_HEAP( afdInfoLength );
    if ( afdInfo == NULL ) {
        return WSAENOBUFS;
    }

    //
    // Set up the AFD information block.
    //

    afdInfo->InformationType = InformationType;

    //
    // If there is additional input information, copy it to the input
    // buffer.
    //

    if ( ARGUMENT_PRESENT( AdditionalInputInfo ) ) {
        WS_ASSERT( AdditionalInputInfoLength != 0 );
        RtlCopyMemory( afdInfo + 1, AdditionalInputInfo, AdditionalInputInfoLength );
    }

    //
    // Set the blocking mode to AFD.
    //

    status = NtDeviceIoControlFile(
                 (HANDLE)Socket->Handle,
                 SockThreadEvent,
                 NULL,                      // APC Routine
                 NULL,                      // APC Context
                 &ioStatusBlock,
                 IOCTL_AFD_GET_INFORMATION,
                 afdInfo,
                 afdInfoLength,
                 afdInfo,
                 sizeof(*afdInfo)
                 );

    //
    // Wait for the operation to complete.
    //

    if ( status == STATUS_PENDING ) {
        SockWaitForSingleObject(
            SockThreadEvent,
            Socket->Handle,
            SOCK_NEVER_CALL_BLOCKING_HOOK,
            SOCK_NO_TIMEOUT
            );
        status = ioStatusBlock.Status;
    }

    if ( !NT_SUCCESS(status) ) {
        FREE_HEAP( afdInfo );
        return SockNtStatusToSocketError( status );
    }

    //
    // Put the return info in the requested parameter.
    //

    if ( ARGUMENT_PRESENT( Boolean ) ) {
        *Boolean = afdInfo->Information.Boolean;
    } else if ( ARGUMENT_PRESENT( Ulong ) ) {
        *Ulong = afdInfo->Information.Ulong;
    } else {
        WS_ASSERT( ARGUMENT_PRESENT( LargeInteger ) );
        *LargeInteger = afdInfo->Information.LargeInteger;
    }

    FREE_HEAP( afdInfo );

    return NO_ERROR;

} // SockGetInformation


INT
SockSetInformation (
    IN PSOCKET_INFORMATION Socket,
    IN ULONG InformationType,
    IN PBOOLEAN Boolean OPTIONAL,
    IN PULONG Ulong OPTIONAL,
    IN PLARGE_INTEGER LargeInteger OPTIONAL
    )
{
    NTSTATUS status;
    IO_STATUS_BLOCK ioStatusBlock;
    AFD_INFORMATION afdInfo;

    //
    // Set up the AFD information block.
    //

    afdInfo.InformationType = InformationType;

    if ( ARGUMENT_PRESENT( Boolean ) ) {
        afdInfo.Information.Boolean = *Boolean;
    } else if ( ARGUMENT_PRESENT( Ulong ) ) {
        afdInfo.Information.Ulong = *Ulong;
    } else {
        ASSERT( ARGUMENT_PRESENT( LargeInteger ) );
        afdInfo.Information.LargeInteger = *LargeInteger;
    }

    //
    // Set the blocking mode to AFD.
    //

    status = NtDeviceIoControlFile(
                 (HANDLE)Socket->Handle,
                 SockThreadEvent,
                 NULL,                      // APC Routine
                 NULL,                      // APC Context
                 &ioStatusBlock,
                 IOCTL_AFD_SET_INFORMATION,
                 &afdInfo,
                 sizeof(afdInfo),
                 NULL,
                 0
                 );

    //
    // Wait for the operation to complete.
    //

    if ( status == STATUS_PENDING ) {
        SockWaitForSingleObject(
            SockThreadEvent,
            Socket->Handle,
            SOCK_NEVER_CALL_BLOCKING_HOOK,
            SOCK_NO_TIMEOUT
            );
        status = ioStatusBlock.Status;
    }

    if ( !NT_SUCCESS(status) ) {
        return SockNtStatusToSocketError( status );
    }

    return NO_ERROR;

} // SockSetInformation

#if DBG
BOOLEAN WsaStartupWarning = FALSE;
#endif


BOOLEAN
SockEnterApi (
    IN BOOLEAN MustBeStarted,
    IN BOOLEAN BlockingIllegal,
    IN BOOLEAN GetXByYCall
    )
{

    //
    // Make sure that WSAStartup has been called, if necessary.
    //

    if ( MustBeStarted && (SockWsaStartupCount == 0 || SockTerminating) ) {

        IF_DEBUG(ENTER) {
            WS_PRINT(( "SockEnterApi: WSAStartup() not called!\n" ));
        }

        SetLastError( WSANOTINITIALISED );
        return FALSE;
    }

    //
    // If this thread has not been initialized, do it now.
    //

    if ( TlsGetValue( SockTlsSlot ) == NULL ) {

        if ( !SockInitialize( NULL, DLL_THREAD_ATTACH, NULL ) ) {

            IF_DEBUG(ENTER) {
                WS_PRINT(( "SockEnterApi: SockInitialize failed.\n" ));
            }

            SetLastError( WSAENOBUFS );
            return FALSE;
        }
    }

    //
    // Make sure that we're not in a blocking call, if appropriate.
    //

    if ( BlockingIllegal && SockThreadIsBlocking ) {

        IF_DEBUG(ENTER) {
            WS_PRINT(( "SockEnterApi: in blocking call.\n" ));
        }

        SetLastError( WSAEINPROGRESS );
        return FALSE;
    }

    //
    // Initialize the cancelled thread variable.  We'll use this to
    // tell whether the operation has been cancelled.
    //

    SockThreadIoCancelled = FALSE;

    //
    // If this is a GetXByY call, set up thread variables.
    //

    if ( GetXByYCall ) {
        SockThreadGetXByYCancelled = FALSE;
        SockThreadProcessingGetXByY = TRUE;
    }

    //
    // Everything's cool.  Proceed.
    //

    return TRUE;

} // SockEnterApi

#if DBG

VOID
WsAssert(
    IN PVOID FailedAssertion,
    IN PVOID FileName,
    IN ULONG LineNumber
    )
{
    BOOL ok;
    CHAR choice[16];
    DWORD bytes;
    DWORD error;

    IF_DEBUG(CONSOLE) {
        WS_PRINT(( "\n failed: %s\n  at line %ld of %s\n",
                    FailedAssertion, LineNumber, FileName ));
        do {
            WS_PRINT(( "[B]reak/[I]gnore? " ));
            bytes = sizeof(choice);
            ok = ReadFile(
                    GetStdHandle(STD_INPUT_HANDLE),
                    &choice,
                    bytes,
                    &bytes,
                    NULL
                    );
            if ( ok ) {
                if ( toupper(choice[0]) == 'I' ) {
                    break;
                }
                if ( toupper(choice[0]) == 'B' ) {
                    DbgUserBreakPoint( );
                }
            } else {
                error = GetLastError( );
            }
        } while ( TRUE );
    
        return;
    }
    
    RtlAssert( FailedAssertion, FileName, LineNumber, NULL );

} // WsAssert

BOOLEAN ConsoleInitialized = FALSE;

HANDLE DebugFileHandle = INVALID_HANDLE_VALUE;
PCHAR DebugFileName = "wsdebug.log";


VOID
WsPrintf (
    char *Format,
    ...
    )

{
    va_list arglist;
    char OutputBuffer[1024];
    ULONG length;
    BOOL ret;

    va_start( arglist, Format );

    vsprintf( OutputBuffer, Format, arglist );

    va_end( arglist );

    IF_DEBUG(DEBUGGER) {
        DbgPrint( OutputBuffer );
    }

    IF_DEBUG(CONSOLE) {

        if ( !ConsoleInitialized ) {
            CONSOLE_SCREEN_BUFFER_INFO csbi;
            COORD coord;
    
            ConsoleInitialized = TRUE;
            (VOID)AllocConsole( );
            (VOID)GetConsoleScreenBufferInfo(
                    GetStdHandle(STD_OUTPUT_HANDLE),
                    &csbi
                    );
            coord.X = (SHORT)(csbi.srWindow.Right - csbi.srWindow.Left + 1);
            coord.Y = (SHORT)((csbi.srWindow.Bottom - csbi.srWindow.Top + 1) * 20);
            (VOID)SetConsoleScreenBufferSize(
                    GetStdHandle(STD_OUTPUT_HANDLE),
                    coord
                    );
        }
    
        length = strlen( OutputBuffer );
    
        ret = WriteFile(
                  GetStdHandle(STD_OUTPUT_HANDLE),
                  (LPVOID )OutputBuffer,
                  length,
                  &length,
                  NULL
                  );
        if ( !ret ) {
            DbgPrint( "WsPrintf: console WriteFile failed: %ld\n",
                          GetLastError( ) );
        }
    }

    IF_DEBUG(FILE) {

        if ( DebugFileHandle == INVALID_HANDLE_VALUE ) {
            DebugFileHandle = CreateFile(
                                  DebugFileName,
                                  GENERIC_READ | GENERIC_WRITE,
                                  FILE_SHARE_READ,
                                  NULL,
                                  CREATE_ALWAYS,
                                  0,
                                  NULL
                                  );
        }

        if ( DebugFileHandle == INVALID_HANDLE_VALUE ) {

            DbgPrint( "WsPrintf: Failed to open winsock debug log file %s: %ld\n",
                          DebugFileName, GetLastError( ) );

        } else {

            length = strlen( OutputBuffer );

            ret = WriteFile(
                      DebugFileHandle,
                      (LPVOID )OutputBuffer,
                      length,
                      &length,
                      NULL
                      );
            if ( !ret ) {
                DbgPrint( "WsPrintf: file WriteFile failed: %ld\n",
                              GetLastError( ) );
            }
        }
    }

} // WsPrintf

#endif


VOID
WsPrintSockaddr (
    IN PSOCKADDR Sockaddr,
    IN PINT SockaddrLength
    )
{

#if DBG

    if ( Sockaddr == NULL ) {
        WS_PRINT(( " NULL addr pointer.\n" ));
        return;
    }

    if ( SockaddrLength == NULL ) {
        WS_PRINT(( " NULL addrlen pointer.\n" ));
        return;
    }

    switch ( Sockaddr->sa_family) {

    case AF_INET: {

        PSOCKADDR_IN sockaddrIn = (PSOCKADDR_IN)Sockaddr;

        if ( *SockaddrLength < sizeof(SOCKADDR_IN) ) {
            WS_PRINT(( " SHORT AF_INET: len %ld\n", *SockaddrLength ));
            return;
        }

        WS_PRINT(( " IP %ld.%ld.%ld.%ld port %ld\n",
                       sockaddrIn->sin_addr.S_un.S_un_b.s_b1,
                       sockaddrIn->sin_addr.S_un.S_un_b.s_b2,
                       sockaddrIn->sin_addr.S_un.S_un_b.s_b3,
                       sockaddrIn->sin_addr.S_un.S_un_b.s_b4,
                       sockaddrIn->sin_port ));
        return;

    }

    default:
        WS_PRINT(( " family %lx\n", Sockaddr->sa_family ));
        break;
    }

    return;

#endif
} // WsPrintSockaddr

#if DBG

VOID
WsEnterApiCall (
    IN PCHAR RoutineName,
    IN PVOID Arg1,
    IN PVOID Arg2,
    IN PVOID Arg3,
    IN PVOID Arg4
    )
{
    ULONG i;

    CHECK_HEAP;

    //
    // If this thread has not been initialized, do it now.  This is
    // duplicated in SockEnterApi(), but we need it here to
    // access SockIndentLevel below.
    //

    if ( TlsGetValue( SockTlsSlot ) == NULL ) {
        if ( !SockInitialize( NULL, DLL_THREAD_ATTACH, NULL ) ) {
            return;
        }
    }

    IF_DEBUG(ENTER) {
        for ( i = 0; i < SockIndentLevel; i++ ) {
            WS_PRINT(( "    " ));
        }
        WS_PRINT(( "---> %s() args 0x%lx 0x%lx 0x%lx 0x%lx\n",
                      RoutineName, Arg1, Arg2, Arg3, Arg4 ));
    }

    SockIndentLevel++;

    return;

} // WsEnter

struct _ERROR_STRINGS {
    INT ErrorCode;
    PCHAR ErrorString;
} ErrorStrings[] = {
    (WSABASEERR+4),   "WSAEINTR",                
    (WSABASEERR+9),   "WSAEBADF",                
    (WSABASEERR+13),  "WSAEACCES",               
    (WSABASEERR+14),  "WSAEFAULT",               
    (WSABASEERR+22),  "WSAEINVAL",               
    (WSABASEERR+24),  "WSAEMFILE",               
    (WSABASEERR+35),  "WSAEWOULDBLOCK",          
    (WSABASEERR+36),  "WSAEINPROGRESS",          
    (WSABASEERR+37),  "WSAEALREADY",             
    (WSABASEERR+38),  "WSAENOTSOCK",             
    (WSABASEERR+39),  "WSAEDESTADDRREQ",         
    (WSABASEERR+40),  "WSAEMSGSIZE",             
    (WSABASEERR+41),  "WSAEPROTOTYPE",           
    (WSABASEERR+42),  "WSAENOPROTOOPT",          
    (WSABASEERR+43),  "WSAEPROTONOSUPPORT",      
    (WSABASEERR+44),  "WSAESOCKTNOSUPPORT",      
    (WSABASEERR+45),  "WSAEOPNOTSUPP",           
    (WSABASEERR+46),  "WSAEPFNOSUPPORT",         
    (WSABASEERR+47),  "WSAEAFNOSUPPORT",         
    (WSABASEERR+48),  "WSAEADDRINUSE",           
    (WSABASEERR+49),  "WSAEADDRNOTAVAIL",        
    (WSABASEERR+50),  "WSAENETDOWN",             
    (WSABASEERR+51),  "WSAENETUNREACH",          
    (WSABASEERR+52),  "WSAENETRESET",            
    (WSABASEERR+53),  "WSAECONNABORTED",         
    (WSABASEERR+54),  "WSAECONNRESET",           
    (WSABASEERR+55),  "WSAENOBUFS",              
    (WSABASEERR+56),  "WSAEISCONN",              
    (WSABASEERR+57),  "WSAENOTCONN",             
    (WSABASEERR+58),  "WSAESHUTDOWN",            
    (WSABASEERR+59),  "WSAETOOMANYREFS",         
    (WSABASEERR+60),  "WSAETIMEDOUT",            
    (WSABASEERR+61),  "WSAECONNREFUSED",         
    (WSABASEERR+62),  "WSAELOOP",                
    (WSABASEERR+63),  "WSAENAMETOOLONG",         
    (WSABASEERR+64),  "WSAEHOSTDOWN",            
    (WSABASEERR+65),  "WSAEHOSTUNREACH",         
    (WSABASEERR+66),  "WSAENOTEMPTY",            
    (WSABASEERR+67),  "WSAEPROCLIM",             
    (WSABASEERR+68),  "WSAEUSERS",               
    (WSABASEERR+69),  "WSAEDQUOT",               
    (WSABASEERR+70),  "WSAESTALE",               
    (WSABASEERR+71),  "WSAEREMOTE",              
    (WSABASEERR+101), "WSAEDISCON",              
    (WSABASEERR+91),  "WSASYSNOTREADY",          
    (WSABASEERR+92),  "WSAVERNOTSUPPORTED",      
    (WSABASEERR+93),  "WSANOTINITIALISED",
    NO_ERROR,         "NO_ERROR"
};


PCHAR
WsGetErrorString (
    IN INT Error
    )
{
    INT i;

    for ( i = 0; ErrorStrings[i].ErrorCode != NO_ERROR; i++ ) {
        if ( ErrorStrings[i].ErrorCode == Error ) {
            return ErrorStrings[i].ErrorString;
        }
    }

    return "Unknown";

} // WsGetErrorString


VOID
WsExitApiCall (
    IN PCHAR RoutineName,
    IN INT ReturnValue,
    IN BOOLEAN Failed

    )
{
    ULONG i;
    INT error = GetLastError( );

    CHECK_HEAP;

    SockIndentLevel--;

    IF_DEBUG(EXIT) {
        for ( i = 0; i < SockIndentLevel; i++ ) {
            WS_PRINT(( "    " ));
        }
        if ( !Failed ) {
            WS_PRINT(( "<--- %s() returning %ld (0x%lx)\n",
                           RoutineName, ReturnValue, ReturnValue ));
        } else {

            PSZ errorString = WsGetErrorString( error );

            WS_PRINT(( "<--- %s() FAILED--error %ld (0x%lx) == %s\n",
                           RoutineName, error, error, errorString ));
        }
    }

    SetLastError( error );

    return;

} // WsExitApiCall

LIST_ENTRY SockHeapListHead;
ULONG SockTotalAllocations = 0;
ULONG SockTotalFrees = 0;
ULONG SockTotalBytesAllocated = 0;
RTL_RESOURCE SocketHeapLock;
BOOLEAN SockHeapDebugInitialized = FALSE;
BOOLEAN SockDebugHeap = FALSE;

PVOID SockHeap = NULL;
PVOID SockCaller1;
PVOID SockCaller2;
BOOLEAN SockDoHeapCheck = FALSE;

#define WINSOCK_HEAP_CODE_1 0xabcdef00
#define WINSOCK_HEAP_CODE_2 0x12345678
#define WINSOCK_HEAP_CODE_3 0x87654321
#define WINSOCK_HEAP_CODE_4 0x00fedcba
#define WINSOCK_HEAP_CODE_5 0xa1b2c3d4

typedef struct _SOCK_HEAP_HEADER {
    ULONG HeapCode1;
    ULONG HeapCode2;
    LIST_ENTRY GlobalHeapListEntry;
    PCHAR FileName;
    ULONG LineNumber;
    ULONG Size;
    ULONG Pad;
} SOCK_HEAP_HEADER, *PSOCK_HEAP_HEADER;

typedef struct _SOCK_HEAP_TAIL {
    PSOCK_HEAP_HEADER Header;
    ULONG HeapCode3;
    ULONG HeapCode4;
    ULONG HeapCode5;
} SOCK_HEAP_TAIL, *PSOCK_HEAP_TAIL;

#define FREE_LIST_SIZE 64
SOCK_HEAP_HEADER SockRecentFreeList[FREE_LIST_SIZE];
ULONG SockRecentFreeListIndex = 0;


VOID
SockInitializeDebugData (
    VOID
    )
{
    RtlInitializeResource( &SocketHeapLock );
    InitializeListHead( &SockHeapListHead );

} // SockInitializeDebugData


PVOID
SockAllocateHeap (
    IN ULONG NumberOfBytes,
    PCHAR FileName,
    ULONG LineNumber
    )
{
    PSOCK_HEAP_HEADER header;
    SOCK_HEAP_TAIL UNALIGNED *tail;

    WS_ASSERT( !SockProcessTerminating );
    WS_ASSERT( (NumberOfBytes & 0xF0000000) == 0 );

    SockCheckHeap( );

    RtlAcquireResourceExclusive( &SocketHeapLock, TRUE );

    header = RtlAllocateHeap( RtlProcessHeap( ), 0,
                              NumberOfBytes + sizeof(*header) + sizeof(*tail) );
    if ( header == NULL ) {
        RtlReleaseResource( &SocketHeapLock );
        return NULL;
    }

    header->HeapCode1 = WINSOCK_HEAP_CODE_1;
    header->HeapCode2 = WINSOCK_HEAP_CODE_2;
    header->FileName = FileName;
    header->LineNumber = LineNumber;
    header->Size = NumberOfBytes;

    tail = (SOCK_HEAP_TAIL UNALIGNED *)( (PCHAR)(header + 1) + NumberOfBytes );
    tail->Header = header;
    tail->HeapCode3 = WINSOCK_HEAP_CODE_3;
    tail->HeapCode4 = WINSOCK_HEAP_CODE_4;
    tail->HeapCode5 = WINSOCK_HEAP_CODE_5;

    InsertTailList( &SockHeapListHead, &header->GlobalHeapListEntry );
    SockTotalAllocations++;
    SockTotalBytesAllocated += header->Size;

    RtlReleaseResource( &SocketHeapLock );

    return (PVOID)(header + 1);

} // SockAllocateHeap


VOID
SockFreeHeap (
    IN PVOID Pointer
    )
{
    PSOCK_HEAP_HEADER header = (PSOCK_HEAP_HEADER)Pointer - 1;
    SOCK_HEAP_TAIL UNALIGNED * tail;

    WS_ASSERT( !SockProcessTerminating );

    SockCheckHeap( );

    tail = (SOCK_HEAP_TAIL UNALIGNED *)( (PCHAR)(header + 1) + header->Size );

    if ( !SockHeapDebugInitialized ) {
        SockInitializeDebugData( );
        SockHeapDebugInitialized = TRUE;
    }

    RtlAcquireResourceExclusive( &SocketHeapLock, TRUE );

    WS_ASSERT( header->HeapCode1 == WINSOCK_HEAP_CODE_1 );
    WS_ASSERT( header->HeapCode2 == WINSOCK_HEAP_CODE_2 );
    WS_ASSERT( tail->HeapCode3 == WINSOCK_HEAP_CODE_3 );
    WS_ASSERT( tail->HeapCode4 == WINSOCK_HEAP_CODE_4 );
    WS_ASSERT( tail->HeapCode5 == WINSOCK_HEAP_CODE_5 );
    WS_ASSERT( tail->Header == header );

    RemoveEntryList( &header->GlobalHeapListEntry );
    SockTotalFrees++;
    SockTotalBytesAllocated -= header->Size;

    //RtlMoveMemory( &SockRecentFreeList[SockRecentFreeListIndex], header, sizeof(*header ) );
    //SockRecentFreeListIndex++;
    //if ( SockRecentFreeListIndex >= FREE_LIST_SIZE ) {
    //    SockRecentFreeListIndex = 0;
    //}

    RtlZeroMemory( header, sizeof(*header) );

    header->HeapCode1 = (ULONG)~WINSOCK_HEAP_CODE_1;
    header->HeapCode2 = (ULONG)~WINSOCK_HEAP_CODE_2;
    tail->HeapCode3 = (ULONG)~WINSOCK_HEAP_CODE_3;
    tail->HeapCode4 = (ULONG)~WINSOCK_HEAP_CODE_4;
    tail->HeapCode5 = (ULONG)~WINSOCK_HEAP_CODE_5;
    tail->Header = NULL;

    RtlReleaseResource( &SocketHeapLock );

    RtlFreeHeap( RtlProcessHeap( ), 0, (PVOID)header );

} // SockFreeHeap


VOID
SockCheckHeap (
    VOID
    )
{
    PLIST_ENTRY listEntry;
    PLIST_ENTRY lastListEntry = NULL;
    PSOCK_HEAP_HEADER header;
    SOCK_HEAP_TAIL UNALIGNED *tail;

    if ( !SockHeapDebugInitialized ) {
        SockInitializeDebugData( );
        SockHeapDebugInitialized = TRUE;
        //SockHeap = RtlCreateHeap( HEAP_GROWABLE, 0, 0, 0, 0, NULL );
        //WS_ASSERT( SockHeap != NULL );
    }

    if ( !SockDoHeapCheck ) {
        return;
    }

    RtlAcquireResourceExclusive( &SocketHeapLock, TRUE );

    for ( listEntry = SockHeapListHead.Flink;
          listEntry != &SockHeapListHead;
          listEntry = listEntry->Flink ) {

        if ( listEntry == NULL ) {
            DbgPrint( "listEntry == NULL, lastListEntry == %lx\n", lastListEntry );
            DbgBreakPoint( );
        }

        header = CONTAINING_RECORD( listEntry, SOCK_HEAP_HEADER, GlobalHeapListEntry );
        tail = (SOCK_HEAP_TAIL UNALIGNED *)( (PCHAR)(header + 1) + header->Size );

        if ( header->HeapCode1 != WINSOCK_HEAP_CODE_1 ) {
            DbgPrint( "SockCheckHeap, fail 1, header %lx tail %lx\n", header, tail );
            DbgBreakPoint( );
        }

        if ( header->HeapCode2 != WINSOCK_HEAP_CODE_2 ) {
            DbgPrint( "SockCheckHeap, fail 2, header %lx tail %lx\n", header, tail );
            DbgBreakPoint( );
        }

        if ( tail->HeapCode3 != WINSOCK_HEAP_CODE_3 ) {
            DbgPrint( "SockCheckHeap, fail 3, header %lx tail %lx\n", header, tail );
            DbgBreakPoint( );
        }

        if ( tail->HeapCode4 != WINSOCK_HEAP_CODE_4 ) {
            DbgPrint( "SockCheckHeap, fail 4, header %lx tail %lx\n", header, tail );
            DbgBreakPoint( );
        }

        if ( tail->HeapCode5 != WINSOCK_HEAP_CODE_5 ) {
            DbgPrint( "SockCheckHeap, fail 5, header %lx tail %lx\n", header, tail );
            DbgBreakPoint( );
        }

        if ( tail->Header != header ) {
            DbgPrint( "SockCheckHeap, fail 6, header %lx tail %lx\n", header, tail );
            DbgBreakPoint( );
        }

        lastListEntry = listEntry;
    }

    RtlGetCallersAddress( &SockCaller1, &SockCaller2 );

    RtlReleaseResource( &SocketHeapLock );

} // SockCheckHeap


VOID
SockAcquireGlobalLockExclusive (
    VOID
    )
{

    WS_ASSERT( !SockProcessTerminating );

    RtlAcquireResourceExclusive( &SocketLock, TRUE );

} // SockAcquireGlobalLockExclusive

VOID
SockAcquireGlobalLockShared (
    VOID
    )
{

    WS_ASSERT( !SockProcessTerminating );

    RtlAcquireResourceShared( &SocketLock, TRUE );

} // SockAcquireGlobalLockShared

VOID
SockAcquireSocketLockExclusive (
    IN PSOCKET_INFORMATION Socket
    )
{

    WS_ASSERT( !SockProcessTerminating );

    RtlAcquireResourceExclusive( &Socket->Lock, TRUE );

} // SockAcquireGlobalLockExclusive

VOID
SockAcquireSocketLockShared (
    IN PSOCKET_INFORMATION Socket
    )
{

    WS_ASSERT( !SockProcessTerminating );

    RtlAcquireResourceShared( &Socket->Lock, TRUE );

} // SockAcquireGlobalLockShared

VOID
SockReleaseGlobalLock (
    VOID
    )
{

    WS_ASSERT( !SockProcessTerminating );

    RtlReleaseResource( &SocketLock );

} // SockReleaseGlobalLock

VOID
SockReleaseSocketLock (
    IN PSOCKET_INFORMATION Socket
    )
{

    WS_ASSERT( !SockProcessTerminating );

    RtlReleaseResource( &Socket->Lock );

} // SockReleaseGlobalLock

#endif // if DBG

